# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import math
import pytest
from picosvg.geometric_types import Rect
from picosvg.svg_transform import Affine2D
from picosvg.svg_types import (
    SVGPath,
    SVGRect,
    SVGLinearGradient,
    SVGRadialGradient,
)
from svg_test_helpers import *


@pytest.mark.parametrize(
    "path, expected_result",
    [
        # path explodes to show implicit commands & becomes absolute
        ("m1,1 2,0 1,3", "M1,1 L3,1 L4,4"),
        # Vertical, Horizontal movement
        ("m2,2 h2 v2 h-1 v-1 H8 V8", "M2,2 H4 V4 H3 V3 H8 V8"),
        # Quadratic bezier curve
        ("m2,2 q1,1 2,2 Q5,5 6,6", "M2,2 Q3,3 4,4 Q5,5 6,6"),
        # Elliptic arc
        ("m2,2 a1,1 0 0 0 3,3 A2,2 1 1 1 4,4", "M2,2 A1 1 0 0 0 5,5 A2 2 1 1 1 4,4"),
        # Cubic bezier
        ("m2,2 c1,-1 2,4 3,3 C4 4 5 5 6 6", "M2,2 C3,1 4,6 5,5 C4,4 5,5 6,6"),
        # Elliptic arc that goes haywire when stroked
        ("M7,5 a3,1 0,0,0 0,-3 a3,3 0 0 1 -4,2", "M7,5 A3 1 0 0 0 7,2 A3 3 0 0 1 3,4"),
        # clock hand's path in which the last point must be == start point when absolutized
        (
            "m63.8 63.98h0.4c2.1 0 3.8-1.7 3.8-3.8v-32.4c0-2.1-1.7-3.8-3.8-3.8h-0.4"
            "c-2.1 0-3.8 1.7-3.8 3.8v32.4c0 2.1 1.7 3.8 3.8 3.8z",
            "M63.8,63.98 H64.2 C66.3,63.98 68,62.28 68,60.18 V27.78 "
            "C68,25.68 66.3,23.98 64.2,23.98 H63.8 C61.7,23.98 60,25.68 60,27.78 "
            "V60.18 C60,62.28 61.7,63.98 63.8,63.98 Z",
        ),
        # Relative 'm' in sub-path following a closed sub-path.
        # Confirms z updates currend position correctly.
        # https://github.com/googlefonts/picosvg/issues/70
        (
            "m0,0 l0,10 l10,0 z m10,10 l0,10 l10,0 z",
            "M0,0 L0,10 L10,10 Z M10,10 L10,20 L20,20 Z",
        ),
        # Further adventures of z; it's a single backref not a stack
        (
            "M3,3 M1,1 l0,10 l4,0 z Z z l8,2 0,2 z m4,4 1,1 -2,0 z",
            "M3,3 M1,1 L1,11 L5,11 Z Z Z L9,3 L9,5 Z M5,5 L6,6 L4,6 Z",
        ),
        # Points very near subpath origin should collapse to that origin, test 1
        # Make sure to test a command with only a single coordinate (h)
        (
            "M0,0 L0,5 L5,5 L1e-10,0 Z l5,-1 0,1 H-1e-9 z",
            "M0,0 L0,5 L5,5 L0,0 Z L5,-1 L5,0 L0,0 Z",
        ),
    ],
)
def test_path_absolute(path: str, expected_result: str):
    actual = SVGPath(d=path).absolute(inplace=True).round_floats(3, inplace=True).d
    print(f"A: {actual}")
    print(f"E: {expected_result}")
    assert actual == expected_result


@pytest.mark.parametrize(
    "path, move, expected_result",
    [
        # path with implicit relative lines
        ("m1,1 2,0 1,3", (2, 2), "M3,3 l2,0 l1,3"),
        # path with implicit absolute lines
        ("M1,1 2,0 1,3", (2, 2), "M3,3 L4,2 L3,5"),
        # Vertical, Horizontal movement
        ("m2,2 h2 v2 h-1 v-1 H8 V8", (-1, -2), "M1,0 h2 v2 h-1 v-1 H7 V6"),
        # Quadratic bezier curve
        ("m2,2 q1,1 2,2 Q5,5 6,6", (3, 1), "M5,3 q1,1 2,2 Q8,6 9,7"),
        # Elliptic arc
        (
            "m2,2 a1,1 0 0 0 3,3 A2,2 1 1 1 4,4",
            (1, 3),
            "M3,5 a1 1 0 0 0 3,3 A2 2 1 1 1 5,7",
        ),
        # Cubic bezier
        ("m2,2 c1,-1 2,4 3,3 C4 4 5 5 6 6", (4, 2), "M6,4 c1,-1 2,4 3,3 C8,6 9,7 10,8"),
    ],
)
def test_path_move(path: str, move, expected_result: str):
    actual = SVGPath(d=path).move(*move, inplace=True).d
    print(f"A: {actual}")
    print(f"E: {expected_result}")
    assert actual == expected_result


@pytest.mark.parametrize(
    "path, expected_result",
    [
        # C/S
        (
            "M600,800 C625,700 725,700 750,800 S875,900 900,800",
            "M600,800 C625,700 725,700 750,800 C775,900 875,900 900,800",
        ),
        # Q/T
        (
            "M16,12 Q20,14 16,16 T16,20 L24,20 24,12",
            "M16,12 Q20,14 16,16 Q12,18 16,20 L24,20 L24,12",
        ),
        # S without preceding C
        ("S875,900 900,800", "C0,0 875,900 900,800"),
        # T without preceding Q
        ("M16,12 T16,20", "M16,12 Q16,12 16,20"),
        # C/s
        (
            "M600,800 C625,700 725,700 750,800 s55,55 200,100",
            "M600,800 C625,700 725,700 750,800 C775,900 805,855 950,900",
        ),
        # https://github.com/rsheeter/playground/issues/4
        (
            "m34 23.25c14.68 0 26.62 18.39 26.62 41s-11.94 41-26.62 41-26.62-18.39-26.62-41 11.94-41 26.62-41z",
            "M34,23.25 c14.68,0 26.62,18.39 26.62,41 C60.62,86.86 48.68,105.25 34,105.25 C19.32,105.25 7.38,86.86 7.38,64.25 C7.38,41.64 19.32,23.25 34,23.25 z",
        ),
    ],
)
def test_expand_shorthand(path, expected_result):
    actual = (
        SVGPath(d=path).expand_shorthand(inplace=True).round_floats(3, inplace=True).d
    )
    print(f"A: {actual}")
    print(f"E: {expected_result}")
    assert actual == expected_result


@pytest.mark.parametrize(
    "shape, expected_bbox",
    [
        # plain rect
        ('<rect x="2" y="2" width="6" height="2" />', Rect(2, 2, 6, 2)),
        # triangle
        ('<path d="m5,2 2.5,5 -5,0 z" />', Rect(2.5, 2, 5, 5)),
    ],
)
def test_bounding_box(shape, expected_bbox):
    nsvg = svg(shape)
    actual_bbox = nsvg.shapes()[0].bounding_box()
    print(f"A: {actual_bbox}")
    print(f"E: {expected_bbox}")
    assert actual_bbox == expected_bbox


@pytest.mark.parametrize(
    "path, expected_result",
    [
        (
            "M-1,0 A1,1 0,0,0 1,0 z",
            "M-1,0 C-1,0.552 -0.552,1 0,1 C0.552,1 1,0.552 1,0 z",
        ),
        # relative coordinates
        (
            "M-1,0 a1,1 0,0,0 2,0 z",
            "M-1,0 C-1,0.552 -0.552,1 0,1 C0.552,1 1,0.552 1,0 z",
        ),
        # degenerate arcs as straight lines
        ("M-1,0 A0,1 0,0,0 0,1 A1,0 0,0,0 1,0 z", "M-1,0 L0,1 L1,0 z"),
        # two arcs closing on each other in a circle; last point == first
        (
            "M4,64 A60 60 0 1 1 124,64 A60 60 0 1 1 4,64 z",
            "M4,64 C4,30.863 30.863,4 64,4 C97.137,4 124,30.863 124,64 "
            "C124,97.137 97.137,124 64,124 C30.863,124 4,97.137 4,64 z",
        ),
    ],
)
def test_arcs_to_cubics(path, expected_result):
    actual = (
        SVGPath(d=path).arcs_to_cubics(inplace=True).round_floats(3, inplace=True).d
    )
    print(f"A: {actual}")
    print(f"E: {expected_result}")
    assert actual == expected_result


@pytest.mark.parametrize(
    "path, transform, expected_result",
    [
        # translate
        (
            "M1,1 L2,1 L2,2 L1,2 Z",
            Affine2D.identity().translate(2, 1),
            "M3,2 L4,2 L4,3 L3,3 Z",
        ),
        # same shape as above under a degenerate transform
        ("M1,1 L2,1 L2,2 L1,2 Z", Affine2D.degenerate(), "M0,0"),
    ],
)
def test_apply_basic_transform(path, transform, expected_result):
    actual = SVGPath(d=path).apply_transform(transform).round_floats(3).d
    print(f"A: {actual}")
    print(f"E: {expected_result}")
    assert actual == expected_result


@pytest.mark.parametrize(
    "path, expected_result",
    [
        (SVGRect(width=1, height=1), True),
        # we see paths with move and nothing else in the wild
        (SVGPath(d="M1,2"), False),
        (SVGPath(d="M1,2 M3,4"), False),
        # a straight line with only a fill (no stroke) and no area does not paint
        (SVGPath(d="M1,2 L3,4 Z"), False),
        # a straight line with a stroke does paint
        (SVGPath(d="M1,2 L3,4", stroke="black"), True),
        # a stroke with 0 width doesn't paint
        (
            SVGPath(d="M1,2 L3,4 L3,1 Z", fill="none", stroke="black", stroke_width=0),
            False,
        ),
        # a filled triangle does paint (no matter the invisible stroke)
        (
            SVGPath(d="M1,2 L3,4 L3,1 Z", fill="red", stroke="black", stroke_width=0),
            True,
        ),
        # we're explicitly told not to display, so we don't
        (SVGPath(d="M1,1 L2,1 L2,2 L1,2 Z", display="none"), False),
        (SVGPath(style="display:none;fill:#F5FAFC;", d="M1,1 L2,1 L2,2 L1,2 Z"), False),
    ],
)
def test_might_paint(path, expected_result):
    assert path.might_paint() == expected_result, path


@pytest.mark.parametrize(
    "shape, expected",
    [
        (
            SVGRect(width=10, height=10, style="fill:red;opacity:0.5;"),
            SVGRect(width=10, height=10, fill="red", opacity=0.5),
        ),
        (
            SVGPath(
                d="M0,0 L10,0 L10,10 L0,10 Z",
                style="stroke:blue;stroke-opacity:0.8;filter:url(#x);",
            ),
            SVGPath(
                d="M0,0 L10,0 L10,10 L0,10 Z",
                stroke="blue",
                stroke_opacity=0.8,
                style="filter:url(#x);",
            ),
        ),
    ],
)
def test_apply_style_attribute(shape, expected):
    actual = shape.apply_style_attribute()
    assert actual == expected
    assert shape.apply_style_attribute(inplace=True) == expected


@pytest.mark.parametrize(
    "path, multiple_of, expected_result",
    [
        ("m1,1 2,0 1,3", 0.1, "m1,1 2,0 1,3"),
        # why a multiple that divides evenly into 1 is a good idea
        ("m1,1 2,0 1,3", 0.128, "m1.024,1.024 2.048,0 1.024,2.944"),
    ],
)
def test_round_multiple(path: str, multiple_of: float, expected_result: str):
    actual = SVGPath(d=path).round_multiple(multiple_of, inplace=True).d
    print(f"A: {actual}")
    print(f"E: {expected_result}")
    assert actual == expected_result


@pytest.mark.parametrize(
    "shape, expected",
    [
        # neither fill nor stroke, unchanged
        (
            SVGPath(d="m1,1 2,0 1,3 z", fill="none", fill_opacity=0.0),
            SVGPath(d="m1,1 2,0 1,3 z", fill="none", fill_opacity=0.0),
        ),
        # both fill and stroke, also unchanged
        (
            SVGPath(
                d="m1,1 2,0 1,3 z",
                fill="red",
                fill_opacity=0.5,
                stroke="black",
                stroke_opacity=0.8,
                opacity=0.9,
            ),
            SVGPath(
                d="m1,1 2,0 1,3 z",
                fill="red",
                fill_opacity=0.5,
                stroke="black",
                stroke_opacity=0.8,
                opacity=0.9,
            ),
        ),
        # no stroke, only fill: merge and drop fill_opacity
        (
            SVGPath(
                d="m1,1 2,0 1,3 z",
                fill="red",
                fill_opacity=0.5,
                opacity=0.9,
            ),
            SVGPath(
                d="m1,1 2,0 1,3 z",
                fill="red",
                opacity=0.45,  # 0.9 * 0.5
            ),
        ),
        # no fill, only stroke: merge and drop stroke_opacity
        (
            SVGPath(
                d="m1,1 2,0 1,3 z",
                fill="none",
                stroke="black",
                stroke_opacity=0.3,
                opacity=0.9,
            ),
            SVGPath(
                d="m1,1 2,0 1,3 z",
                fill="none",
                stroke="black",
                opacity=0.27,  # 0.9 * 0.3
            ),
        ),
    ],
)
def test_normalize_opacity(shape, expected):
    assert shape.normalize_opacity() == expected


@pytest.mark.parametrize(
    "el, view_box, expected",
    [
        # default linearGradient
        (
            etree.Element("linearGradient", {"id": "lg1"}),
            Rect(0, 0, 10, 10),
            SVGLinearGradient(
                id="lg1",
                x1=0.0,
                y1=0.0,
                x2=1.0,
                y2=0.0,
                gradientUnits="objectBoundingBox",
            ),
        ),
        # default radialGradient
        (
            etree.Element("radialGradient", {"id": "rg2"}),
            Rect(0, 0, 10, 10),
            SVGRadialGradient(
                id="rg2",
                cx=0.5,
                cy=0.5,
                r=0.5,
                fx=0.5,
                fy=0.5,
                fr=0.0,
                gradientUnits="objectBoundingBox",
            ),
        ),
        # radialGradient with gradientUnits="userSpaceOnUse" on square viewport
        (
            etree.Element(
                "radialGradient", {"id": "rg3", "gradientUnits": "userSpaceOnUse"}
            ),
            Rect(0, 0, 10, 10),
            SVGRadialGradient(
                id="rg3",
                cx=5.0,
                cy=5.0,
                r=5.0,
                fx=5.0,
                fy=5.0,
                fr=0.0,
                gradientUnits="userSpaceOnUse",
            ),
        ),
        # userSpaceOnUse & nonsquare viewport, default 'r' is 50% of normalized diagonal
        (
            etree.Element(
                "radialGradient", {"id": "rg4", "gradientUnits": "userSpaceOnUse"}
            ),
            Rect(0, 0, 10, 5),
            SVGRadialGradient(
                id="rg4",
                cx=5.0,
                cy=2.5,
                r=0.5 * math.hypot(10, 5) / math.sqrt(2),  # 3.952847
                fx=5.0,
                fy=2.5,
                fr=0.0,
                gradientUnits="userSpaceOnUse",
            ),
        ),
        # fx/fy default to cx/cy when not explicitly set
        (
            etree.Element("radialGradient", {"id": "rg5", "cx": "20%", "cy": "40%"}),
            Rect(0, 0, 10, 5),
            SVGRadialGradient(
                id="rg5",
                cx=0.2,
                cy=0.4,
                r=0.5,
                fx=0.2,
                fy=0.4,
                fr=0.0,
                gradientUnits="objectBoundingBox",
            ),
        ),
        # fx/fy explicitly set
        (
            etree.Element("radialGradient", {"id": "rg6", "fx": "20%", "fy": "40%"}),
            Rect(0, 0, 10, 5),
            SVGRadialGradient(
                id="rg6",
                cx=0.5,
                cy=0.5,
                r=0.5,
                fx=0.2,
                fy=0.4,
                fr=0.0,
                gradientUnits="objectBoundingBox",
            ),
        ),
        # linearGradient with gradientTransform and spreadMethod
        (
            etree.Element(
                "linearGradient",
                {
                    "id": "lg7",
                    "gradientTransform": "matrix(1, 0.3, 0, 1, 0, 0)",
                    "spreadMethod": "reflect",
                },
            ),
            Rect(0, 0, 10, 10),
            SVGLinearGradient(
                id="lg7",
                x1=0.0,
                y1=0.0,
                x2=1.0,
                y2=0.0,
                gradientTransform=Affine2D(1, 0.3, 0, 1, 0, 0),
                spreadMethod="reflect",
            ),
        ),
    ],
)
def test_gradient_from_element(el, view_box, expected):
    klass = type(expected)
    assert klass.from_element(el, view_box) == expected



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import dataclasses
from textwrap import dedent
from lxml import etree
import os
import pytest
from picosvg.svg import SVG, SVGPath
from picosvg.svg_meta import strip_ns, parse_css_declarations
from svg_test_helpers import *


def _test(actual, expected_result, op):
    actual = op(load_test_svg(actual))
    expected_result = load_test_svg(expected_result)
    drop_whitespace(actual)
    drop_whitespace(expected_result)
    print(f"A: {pretty_print(actual.toetree())}")
    print(f"E: {pretty_print(expected_result.toetree())}")
    assert pretty_print(actual.toetree()) == pretty_print(expected_result.toetree())


@pytest.mark.parametrize(
    "shape, expected_fields",
    [
        # path, fill
        ("<path d='M1,1 2,2' fill='blue' />", {"fill": "blue"}),
        # rect, opacity
        ("<rect x='5' y='5' width='5' height='5' opacity='0.5'/>", {"opacity": 0.5}),
        # polyline, clip-path
        (
            "<polyline points='1,1 5,5 2,2' clip-path='url(#cp)'/>",
            {"clip_path": "url(#cp)"},
        ),
        # line, stroke
        ("<line x1='1' y1='1' x2='10' y2='10' stroke='red'/>", {"stroke": "red"}),
    ],
)
def test_common_attrib(shape, expected_fields):
    svg = SVG.fromstring(svg_string(shape))
    field_values = dataclasses.asdict(svg.shapes()[0])
    for field_name, expected_value in expected_fields.items():
        assert field_values.get(field_name, "") == expected_value, field_name

    svg = svg.shapes_to_paths()
    field_values = dataclasses.asdict(svg.shapes()[0])
    for field_name, expected_value in expected_fields.items():
        assert field_values.get(field_name, "") == expected_value, field_name


# https://www.w3.org/TR/SVG11/shapes.html
@pytest.mark.parametrize(
    "shape, expected_path",
    [
        # path: direct passthrough
        ("<path d='I love kittens'/>", 'd="I love kittens"'),
        # path no @d
        ("<path duck='Mallard'/>", ""),
        # line
        ('<line x1="10" x2="50" y1="110" y2="150"/>', 'd="M10,110 L50,150"'),
        # line, decimal positioning
        (
            '<line x1="10.0" x2="50.5" y1="110.2" y2="150.7"/>',
            'd="M10,110.2 L50.5,150.7"',
        ),
        # rect: minimal valid example
        ("<rect width='1' height='1'/>", 'd="M0,0 H1 V1 H0 V0 Z"'),
        # rect: sharp corners
        (
            "<rect x='10' y='11' width='17' height='11'/>",
            'd="M10,11 H27 V22 H10 V11 Z"',
        ),
        # rect: round corners
        (
            "<rect x='9' y='9' width='11' height='7' rx='2'/>",
            'd="M11,9 H18 A2 2 0 0 1 20,11 V14 A2 2 0 0 1 18,16 H11'
            ' A2 2 0 0 1 9,14 V11 A2 2 0 0 1 11,9 Z"',
        ),
        # rect: simple
        (
            "<rect x='11.5' y='16' width='11' height='2'/>",
            'd="M11.5,16 H22.5 V18 H11.5 V16 Z"',
        ),
        # polygon
        ("<polygon points='30,10 50,30 10,30'/>", 'd="M30,10 50,30 10,30 Z"'),
        # polyline
        ("<polyline points='30,10 50,30 10,30'/>", 'd="M30,10 50,30 10,30"'),
        # circle, minimal valid example
        ("<circle r='1'/>", 'd="M1,0 A1 1 0 1 1 -1,0 A1 1 0 1 1 1,0 Z"'),
        # circle
        (
            "<circle cx='600' cy='200' r='100'/>",
            'd="M700,200 A100 100 0 1 1 500,200 A100 100 0 1 1 700,200 Z"',
        ),
        # circle, decimal positioning
        (
            "<circle cx='12' cy='6.5' r='1.5'></circle>",
            'd="M13.5,6.5 A1.5 1.5 0 1 1 10.5,6.5 A1.5 1.5 0 1 1 13.5,6.5 Z"',
        ),
        # ellipse
        (
            '<ellipse cx="100" cy="50" rx="100" ry="50"/>',
            'd="M200,50 A100 50 0 1 1 0,50 A100 50 0 1 1 200,50 Z"',
        ),
        # ellipse, decimal positioning
        (
            '<ellipse cx="100.5" cy="50" rx="10" ry="50.5"/>',
            'd="M110.5,50 A10 50.5 0 1 1 90.5,50 A10 50.5 0 1 1 110.5,50 Z"',
        ),
    ],
)
def test_shapes_to_paths(shape: str, expected_path: str):
    actual = SVG.fromstring(svg_string(shape)).shapes_to_paths(inplace=True).toetree()
    expected_result = SVG.fromstring(svg_string(f"<path {expected_path}/>")).toetree()
    print(f"A: {pretty_print(actual)}")
    print(f"E: {pretty_print(expected_result)}")
    assert etree.tostring(actual) == etree.tostring(expected_result)


@pytest.mark.parametrize(
    "shape, expected_cmds",
    [
        # line
        (
            '<line x1="10" x2="50" y1="110" y2="150"/>',
            [("M", (10.0, 110.0)), ("L", (50.0, 150.0))],
        ),
        # path explodes to show implicit commands
        (
            '<path d="m1,1 2,0 1,3"/>',
            [("m", (1.0, 1.0)), ("l", (2.0, 0.0)), ("l", (1.0, 3.0))],
        ),
        # vertical and horizontal movement
        (
            '<path d="m1,1 v2 h2z"/>',
            [("m", (1.0, 1.0)), ("v", (2.0,)), ("h", (2.0,)), ("z", ())],
        ),
        # arc, negative offsets
        (
            '<path d="M7,5 a3,1 0,0,0 0,-3 a3,3 0 0 1 -4,2"/>',
            [
                ("M", (7.0, 5.0)),
                ("a", (3.0, 1.0, 0.0, 0.0, 0.0, 0.0, -3.0)),
                ("a", (3.0, 3.0, 0.0, 0.0, 1.0, -4.0, 2.0)),
            ],
        ),
        # minimalist numbers, who needs spaces or commas
        (
            '<path d="m-1-1 0.5-.5-.5-.3.1.2.2.51.52.711"/>',
            [
                ("m", (-1.0, -1.0)),
                ("l", (0.5, -0.5)),
                ("l", (-0.5, -0.3)),
                ("l", (0.1, 0.2)),
                ("l", (0.2, 0.51)),
                ("l", (0.52, 0.711)),
            ],
        ),
    ],
)
def test_iter(shape, expected_cmds):
    svg_path = SVG.fromstring(svg_string(shape)).shapes_to_paths().shapes()[0]
    actual_cmds = [t for t in svg_path]
    print(f"A: {actual_cmds}")
    print(f"E: {expected_cmds}")
    assert actual_cmds == expected_cmds


@pytest.mark.parametrize(
    "actual, expected_result", [("use-ellipse.svg", "use-ellipse-resolved.svg")]
)
def test_resolve_use(actual, expected_result):
    _test(actual, expected_result, lambda svg: svg.resolve_use(inplace=True))


@pytest.mark.parametrize(
    "actual, expected_result",
    [
        ("stroke-simplepath-before.svg", "stroke-simplepath-nano.svg"),
        ("stroke-path-before.svg", "stroke-path-nano.svg"),
        ("stroke-capjoinmiterlimit-before.svg", "stroke-capjoinmiterlimit-nano.svg"),
        ("scale-strokes-before.svg", "scale-strokes-nano.svg"),
        ("stroke-fill-opacity-before.svg", "stroke-fill-opacity-nano.svg"),
        ("stroke-dasharray-before.svg", "stroke-dasharray-nano.svg"),
        ("stroke-circle-dasharray-before.svg", "stroke-circle-dasharray-nano.svg"),
        ("clip-rect.svg", "clip-rect-clipped-nano.svg"),
        ("clip-ellipse.svg", "clip-ellipse-clipped-nano.svg"),
        ("clip-curves.svg", "clip-curves-clipped-nano.svg"),
        ("clip-multirect.svg", "clip-multirect-clipped-nano.svg"),
        ("clip-groups.svg", "clip-groups-clipped-nano.svg"),
        ("clip-use.svg", "clip-use-clipped-nano.svg"),
        ("clip-rule-example.svg", "clip-rule-example-nano.svg"),
        ("clip-from-brazil-flag.svg", "clip-from-brazil-flag-nano.svg"),
        ("clip-rule-evenodd.svg", "clip-rule-evenodd-clipped-nano.svg"),
        ("clip-clippath-attrs.svg", "clip-clippath-attrs-nano.svg"),
        ("clip-clippath-none.svg", "clip-clippath-none-nano.svg"),
        ("rotated-rect.svg", "rotated-rect-nano.svg"),
        ("translate-rect.svg", "translate-rect-nano.svg"),
        ("ungroup-before.svg", "ungroup-nano.svg"),
        ("ungroup-multiple-children-before.svg", "ungroup-multiple-children-nano.svg"),
        ("group-stroke-before.svg", "group-stroke-nano.svg"),
        ("arcs-before.svg", "arcs-nano.svg"),
        ("invisible-before.svg", "invisible-nano.svg"),
        ("transform-before.svg", "transform-nano.svg"),
        ("group-data-name-before.svg", "group-data-name-nano.svg"),
        ("matrix-before.svg", "matrix-nano.svg"),
        ("degenerate-before.svg", "degenerate-nano.svg"),
        ("fill-rule-evenodd-before.svg", "fill-rule-evenodd-nano.svg"),
        ("twemoji-lesotho-flag-before.svg", "twemoji-lesotho-flag-nano.svg"),
        ("inline-css-style-before.svg", "inline-css-style-nano.svg"),
        ("clipped-strokes-before.svg", "clipped-strokes-nano.svg"),
        ("drop-anon-symbols-before.svg", "drop-anon-symbols-nano.svg"),
        ("scale-strokes-before.svg", "scale-strokes-nano.svg"),
        ("ungroup-with-ids-before.svg", "ungroup-with-ids-nano.svg"),
        ("stroke-with-id-before.svg", "stroke-with-id-nano.svg"),
        ("drop-title-meta-desc-before.svg", "drop-title-meta-desc-nano.svg"),
        ("no-viewbox-before.svg", "no-viewbox-nano.svg"),
        ("decimal-viewbox-before.svg", "decimal-viewbox-nano.svg"),
        ("inkscape-noise-before.svg", "inkscape-noise-nano.svg"),
        ("flag-use-before.svg", "flag-use-nano.svg"),
        ("ungroup-transform-before.svg", "ungroup-transform-nano.svg"),
        ("pathops-tricky-path-before.svg", "pathops-tricky-path-nano.svg"),
        ("gradient-template-1-before.svg", "gradient-template-1-nano.svg"),
        ("nested-svg-slovenian-flag-before.svg", "nested-svg-slovenian-flag-nano.svg"),
        ("global-fill-none-before.svg", "global-fill-none-nano.svg"),
        ("stroke-polyline-before.svg", "stroke-polyline-nano.svg"),
        ("clip-the-clip-before.svg", "clip-the-clip-nano.svg"),
        ("ungroup-group-transform-before.svg", "ungroup-group-transform-nano.svg"),
        ("ungroup-transform-clip-before.svg", "ungroup-transform-clip-nano.svg"),
        (
            "ungroup-retain-for-opacity-before.svg",
            "ungroup-retain-for-opacity-nano.svg",
        ),
        (
            "transform-radial-userspaceonuse-before.svg",
            "transform-radial-userspaceonuse-nano.svg",
        ),
        (
            "transform-linear-objectbbox-before.svg",
            "transform-linear-objectbbox-nano.svg",
        ),
        (
            "transform-radial-objectbbox-before.svg",
            "transform-radial-objectbbox-nano.svg",
        ),
        (
            "illegal-inheritance-before.svg",
            "illegal-inheritance-nano.svg",
        ),
        (
            "explicit-default-fill-no-inherit-before.svg",
            "explicit-default-fill-no-inherit-nano.svg",
        ),
        (
            "explicit-default-stroke-no-inherit-before.svg",
            "explicit-default-stroke-no-inherit-nano.svg",
        ),
        (
            "inherit-default-fill-before.svg",
            "inherit-default-fill-nano.svg",
        ),
    ],
)
def test_topicosvg(actual, expected_result):
    _test(actual, expected_result, lambda svg: svg.topicosvg())


@pytest.mark.parametrize(
    "actual, expected_result",
    [
        ("outside-viewbox.svg", "outside-viewbox-clipped.svg"),
        ("outside-viewbox-grouped.svg", "outside-viewbox-grouped-clipped.svg"),
    ],
)
def test_clip_to_viewbox(actual, expected_result):
    _test(actual, expected_result, lambda svg: svg.clip_to_viewbox().round_floats(4))


@pytest.mark.parametrize(
    "actual, expected_result", [("invisible-before.svg", "invisible-after.svg")]
)
def test_remove_unpainted_shapes(actual, expected_result):
    _test(actual, expected_result, lambda svg: svg.remove_unpainted_shapes())


@pytest.mark.parametrize(
    "svg_file, expected_violations",
    [
        ("good-defs-0.svg", ()),
        (
            "bad-defs-0.svg",
            (
                "BadElement: /svg[0]/defs[1]",
                "BadElement: /svg[0]/donkey[0]",
            ),
        ),
        ("bad-defs-1.svg", ("MissingElement: /svg[0]/defs[0]",)),
        (
            "bad-ids-1.svg",
            (
                'BadElement: /svg[0]/path[1] reuses id="not_so_unique", first seen at /svg[0]/path[0]',
            ),
        ),
    ],
)
def test_checkpicosvg(svg_file, expected_violations):
    nano_violations = load_test_svg(svg_file).checkpicosvg()
    assert expected_violations == nano_violations


@pytest.mark.parametrize(
    "svg_string, expected_result",
    [
        ('<svg version="1.1" xmlns="http://www.w3.org/2000/svg"/>', None),
        (
            '<svg version="1.1" xmlns="http://www.w3.org/2000/svg" viewBox="7 7 12 12"/>',
            (7, 7, 12, 12),
        ),
    ],
)
def test_viewbox(svg_string, expected_result):
    assert SVG.fromstring(svg_string).view_box() == expected_result


@pytest.mark.parametrize(
    "svg_string, names, expected_result",
    [
        # No change
        (
            '<svg xmlns="http://www.w3.org/2000/svg" version="1.1"/>',
            ("viewBox", "width", "height"),
            '<svg xmlns="http://www.w3.org/2000/svg" version="1.1"/>',
        ),
        # Drop viewBox, width, height
        (
            '<svg xmlns="http://www.w3.org/2000/svg" version="1.1" viewBox="7 7 12 12" height="7" width="11"/>',
            ("viewBox", "width", "height"),
            '<svg xmlns="http://www.w3.org/2000/svg" version="1.1"/>',
        ),
        # Drop width, height
        (
            '<svg xmlns="http://www.w3.org/2000/svg" version="1.1" viewBox="7 7 12 12" height="7" width="11"/>',
            ("width", "height"),
            '<svg xmlns="http://www.w3.org/2000/svg" version="1.1" viewBox="7 7 12 12"/>',
        ),
    ],
)
def test_remove_attributes(svg_string, names, expected_result):
    assert (
        SVG.fromstring(svg_string).remove_attributes(names).tostring()
    ) == expected_result


# https://github.com/rsheeter/picosvg/issues/1
@pytest.mark.parametrize(
    "svg_string, expected_result",
    [
        (
            '<svg version="1.1" xmlns="http://www.w3.org/2000/svg" viewBox="7 7 12 12"/>',
            0.012,
        ),
        (
            '<svg version="1.1" xmlns="http://www.w3.org/2000/svg" viewBox="0 0 128 128"/>',
            0.128,
        ),
    ],
)
def test_tolerance(svg_string, expected_result):
    assert round(SVG.fromstring(svg_string).tolerance, 4) == expected_result


@pytest.mark.parametrize(
    "style, property_names, expected_output, expected_unparsed",
    [
        ("fill:none", None, {"fill": "none"}, ""),
        ("fill: url(#grad1)", None, {"fill": "url(#grad1)"}, ""),
        (
            " stroke  : blue   ; stroke-width :4;   ",
            None,
            {"stroke": "blue", "stroke-width": "4"},
            "",
        ),
        (
            "enable-background:new 0 0 128 128; foo:abc; bar:123;",
            {"enable-background"},
            {"enable-background": "new 0 0 128 128"},
            "foo:abc; bar:123;",
        ),
    ],
)
def test_parse_css_declarations(
    style, property_names, expected_output, expected_unparsed
):
    output = {}
    unparsed = parse_css_declarations(style, output, property_names)
    assert output == expected_output
    assert unparsed == expected_unparsed


@pytest.mark.parametrize("style", ["foo;bar;", "foo:bar:baz;"])
def test_parse_css_declarations_invalid(style):
    with pytest.raises(ValueError, match="Invalid CSS declaration syntax"):
        parse_css_declarations(style, {})


@pytest.mark.parametrize(
    "actual, expected_result",
    [("inline-css-style-before.svg", "inline-css-style-after.svg")],
)
def test_apply_style_attributes(actual, expected_result):
    _test(actual, expected_result, lambda svg: svg.apply_style_attributes())
    # check we get the same output even if shapes were already parsed
    _test(
        actual,
        expected_result,
        lambda svg: svg.shapes() and svg.apply_style_attributes(),
    )


@pytest.mark.parametrize(
    "gradient_string, expected_result",
    [
        # No transform, no change
        (
            '<linearGradient id="c" x1="63.85" x2="63.85" y1="4245" y2="4137.3" gradientUnits="userSpaceOnUse"/>',
            '<linearGradient id="c" x1="63.85" y1="4245" x2="63.85" y2="4137.3" gradientUnits="userSpaceOnUse"/>',
        ),
        # Real example from emoji_u1f392.svg w/ dx changed from 0 to 1
        # scale, translate
        (
            '<linearGradient id="c" x1="63.85" x2="63.85" y1="4245" y2="4137.3" gradientTransform="translate(1 -4122)" gradientUnits="userSpaceOnUse"/>',
            '<linearGradient id="c" x1="64.85" y1="123" x2="64.85" y2="15.3" gradientUnits="userSpaceOnUse"/>',
        ),
        # Real example from emoji_u1f392.svg w/sx changed from 1 to 0.5
        # scale, translate
        (
            '<radialGradient id="b" cx="63.523" cy="12368" r="53.477" gradientTransform="matrix(.5 0 0 .2631 0 -3150)" gradientUnits="userSpaceOnUse"/>',
            '<radialGradient id="b" cx="63.523" cy="395.366021" r="53.477" gradientTransform="matrix(0.5 0 0 0.2631 0 0)" gradientUnits="userSpaceOnUse"/>',
        ),
        # Real example from emoji_u1f44d.svg
        # Using all 6 parts
        (
            '<radialGradient id="d" cx="2459.4" cy="-319.18" r="20.331" gradientTransform="matrix(-1.3883 .0794 -.0374 -.6794 3505.4 -353.39)" gradientUnits="userSpaceOnUse"/>',
            '<radialGradient id="d" cx="-71.60264" cy="-94.82264" r="20.331" gradientTransform="matrix(-1.3883 0.0794 -0.0374 -0.6794 0 0)" gradientUnits="userSpaceOnUse"/>',
        ),
        # Manually constructed objectBBox
        (
            '<radialGradient id="mbbox" cx="0.75" cy="0.75" r="0.40" gradientTransform="matrix(1 1 -0.7873 -0.001717 0.5 0)" gradientUnits="objectBoundingBox"/>',
            '<radialGradient id="mbbox" cx="0.748907" cy="0.11353" r="0.4" gradientTransform="matrix(1 1 -0.7873 -0.001717 0 0)"/>',
        ),
        # Real example from emoji_u26BE
        # https://github.com/googlefonts/picosvg/issues/129
        (
            '<radialGradient id="f" cx="-779.79" cy="3150" r="58.471" gradientTransform="matrix(0 1 -1 0 3082.5 1129.5)" gradientUnits="userSpaceOnUse"/>',
            '<radialGradient id="f" cx="349.71" cy="67.5" r="58.471" gradientTransform="matrix(0 1 -1 0 0 0)" gradientUnits="userSpaceOnUse"/>',
        ),
        # Real example from emoji_u270c.svg
        # Very small values (e-17...) and float math makes for large errors
        (
            '<radialGradient id="f" cx="75.915" cy="20.049" r="71.484" fx="88.617" fy="-50.297" gradientTransform="matrix(6.1232e-17 1 -1.0519 6.4408e-17 97.004 -55.866)" gradientUnits="userSpaceOnUse"/>',
            '<radialGradient id="f" cx="20.049" cy="-72.168891" r="71.484" fx="32.751" fy="-142.514891" gradientTransform="matrix(0 1 -1.0519 0 0 0)" gradientUnits="userSpaceOnUse"/>',
        ),
    ],
)
def test_apply_gradient_translation(gradient_string, expected_result):
    svg = SVG.fromstring(svg_string(gradient_string))
    for grad_el in svg._select_gradients():
        svg._apply_gradient_translation(grad_el)
    el = svg.xpath_one("//svg:linearGradient | //svg:radialGradient")

    for node in svg.svg_root.getiterator():
        node.tag = etree.QName(node).localname
    etree.cleanup_namespaces(svg.svg_root)

    assert etree.tostring(el).decode("utf-8") == expected_result


@pytest.mark.parametrize(
    "svg_content, expected_result",
    [
        # Blank fill
        # https://github.com/googlefonts/nanoemoji/issues/229
        (
            '<path fill="" d=""/>',
            (SVGPath(),),
        ),
    ],
)
def test_default_for_blank(svg_content, expected_result):
    assert tuple(SVG.fromstring(svg_string(svg_content)).shapes()) == expected_result


@pytest.mark.parametrize(
    "actual, expected_result",
    [
        ("gradient-template-1-before.svg", "gradient-template-1-after.svg"),
        ("gradient-template-2-before.svg", "gradient-template-2-after.svg"),
        ("gradient-template-3-before.svg", "gradient-template-3-after.svg"),
    ],
)
def test_resolve_gradient_templates(actual, expected_result):
    def apply_templates(svg):
        for grad_el in svg._select_gradients():
            svg._apply_gradient_template(grad_el)
        svg._remove_orphaned_gradients()
        return svg

    _test(
        actual,
        expected_result,
        apply_templates,
    )


@pytest.mark.parametrize(
    "actual, expected_result",
    [
        ("nested-svg-slovenian-flag-before.svg", "nested-svg-slovenian-flag-after.svg"),
    ],
)
def test_resolve_nested_svgs(actual, expected_result):
    _test(
        actual,
        expected_result,
        lambda svg: svg.resolve_nested_svgs(),
    )


def test_tostring_pretty_print():
    svg = SVG.fromstring(
        '<svg xmlns="http://www.w3.org/2000/svg" version="1.1" viewBox="0 0 128 128">\n'
        "<g>  \n"
        "\t  <g>  \r\n"
        '\t\t  <path d="M60,30 L100,30 L100,70 L60,70 Z"/>\n\n'
        "\t  </g>  \r"
        "</g> \n"
        "</svg>"
    )

    assert svg.tostring(pretty_print=False) == (
        '<svg xmlns="http://www.w3.org/2000/svg" version="1.1" viewBox="0 0 128 128">'
        "<g>"
        "<g>"
        '<path d="M60,30 L100,30 L100,70 L60,70 Z"/>'
        "</g>"
        "</g>"
        "</svg>"
    )

    assert svg.tostring(pretty_print=True) == dedent(
        """\
        <svg xmlns="http://www.w3.org/2000/svg" version="1.1" viewBox="0 0 128 128">
          <g>
            <g>
              <path d="M60,30 L100,30 L100,70 L60,70 Z"/>
            </g>
          </g>
        </svg>
        """
    )


@pytest.mark.parametrize(
    "actual, expected_result",
    [
        ("fill-rule-evenodd-before.svg", "fill-rule-evenodd-after.svg"),
    ],
)
def test_evenodd_to_nonzero_winding(actual, expected_result):
    _test(
        actual,
        expected_result,
        lambda svg: svg.evenodd_to_nonzero_winding().round_floats(3, inplace=True),
    )


@pytest.mark.parametrize(
    "input_svg",
    (
        "explicit-default-fill-no-inherit-before.svg",
        "explicit-default-stroke-no-inherit-before.svg",
        "inherit-default-fill-before.svg",
    ),
)
def test_update_tree_lossless(input_svg):
    with open(locate_test_file(input_svg)) as f:
        svg_data = f.read()
    svg = SVG.fromstring(svg_data)
    assert not svg.elements  # initially empty list

    # _elements() parses shapes using from_element, populating self.elements
    _ = svg._elements()
    assert svg.elements

    # _update_etree calls to_element on each shape and resets self.elements
    svg._update_etree()
    assert not svg.elements

    assert svg.tostring(pretty_print=True) == svg_data



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Functions meant to help discern if shapes are the same or not.

import copy
import dataclasses
from itertools import islice
from math import atan2
from picosvg.geometric_types import Vector, almost_equal
from picosvg.svg_types import SVGShape, SVGPath
from typing import Callable, Generator, Iterable, Optional, Tuple
from picosvg import svg_meta
from picosvg.svg_transform import Affine2D


_SIGNIFICANCE_FACTOR = 5  # Must be at least N x tolerance to be significant
_ROUND_RANGE = range(3, 13)  # range of rounds to try


def _first_move(path: SVGPath) -> Tuple[float, float]:
    cmd, args = next(iter(path))
    if cmd.upper() != "M":
        raise ValueError(f"Path for {path} should start with a move")
    return args


def _vectors(path: SVGPath) -> Generator[Vector, None, None]:
    for cmd, args in path:
        x_coord_idxs, y_coord_idxs = svg_meta.cmd_coords(cmd)
        if cmd.lower() == "z":
            yield Vector(0.0, 0.0)
        else:
            yield Vector(args[x_coord_idxs[-1]], args[y_coord_idxs[-1]])


def _nth_vector(path: SVGPath, n: int) -> Vector:
    return next(islice(_vectors(path), n, n + 1))


def _angle(v: Vector) -> float:
    # gives the directional angle of vector (unlike acos)
    return atan2(v.y, v.x)


def _affine_vec2vec(initial: Vector, target: Vector) -> Affine2D:
    affine = Affine2D.identity()

    # rotate initial to have the same angle as target (may have different magnitude)
    angle = _angle(target) - _angle(initial)
    affine = Affine2D.identity().rotate(angle)
    vec = affine.map_vector(initial)

    # scale to target magnitude
    s = 0
    if vec.norm() != 0:
        s = target.norm() / vec.norm()

    affine = Affine2D.compose_ltr((affine, Affine2D.identity().scale(s, s)))

    return affine


def _first_significant(
    vectors: Iterable[Vector], val_fn: Callable[[Vector], float], tolerance: float
) -> Tuple[int, Optional[Vector]]:
    tolerance = _SIGNIFICANCE_FACTOR * tolerance
    for idx, vec in enumerate(vectors):
        if idx == 0:  # skip initial move
            continue
        if abs(val_fn(vec)) > tolerance:
            return (idx, vec)
    return (-1, None)


def _first_significant_for_both(
    s1: SVGPath, s2: SVGPath, val_fn: Callable[[Vector], float], tolerance: float
) -> Tuple[int, Optional[Vector], Optional[Vector]]:
    tolerance = _SIGNIFICANCE_FACTOR * tolerance
    for idx, (vec1, vec2) in enumerate(zip(_vectors(s1), _vectors(s2))):
        if idx == 0:  # skip initial move
            continue
        if abs(val_fn(vec1)) > tolerance and abs(val_fn(vec2)) > tolerance:
            return (idx, vec1, vec2)
    return (-1, None, None)


# Makes a shape safe for a walk with _affine_callback
def _affine_friendly(shape: SVGShape) -> SVGPath:
    path = shape.as_path()
    if shape is path:
        path = copy.deepcopy(path)
    return (
        path.relative(inplace=True)
        .explicit_lines(inplace=True)
        .expand_shorthand(inplace=True)
    )


# Transform all coords in an affine-friendly path
def _affine_callback(affine, subpath_start, curr_pos, cmd, args, *_unused):
    x_coord_idxs, y_coord_idxs = svg_meta.cmd_coords(cmd)
    # hard to do things like rotate if we have 1d coords
    assert len(x_coord_idxs) == len(y_coord_idxs), f"{cmd}, {args}"

    args = list(args)  # we'd like to mutate 'em
    for x_coord_idx, y_coord_idx in zip(x_coord_idxs, y_coord_idxs):
        if cmd == cmd.upper():
            # for an absolute cmd allow translation: map_point
            new_x, new_y = affine.map_point((args[x_coord_idx], args[y_coord_idx]))
        else:
            # for a relative coord no translate: map_vector
            new_x, new_y = affine.map_vector((args[x_coord_idx], args[y_coord_idx]))

        if almost_equal(new_x, 0):
            new_x = 0
        if almost_equal(new_y, 0):
            new_y = 0
        args[x_coord_idx] = new_x
        args[y_coord_idx] = new_y
    return ((cmd, args),)


def normalize(shape: SVGShape, tolerance: float) -> SVGPath:
    """Build a version of shape that will compare == to other shapes even if offset,
    scaled, rotated, etc.

    Intended use is to normalize multiple shapes to identify opportunity for reuse."""

    path = _affine_friendly(dataclasses.replace(shape, id=""))

    # Make path relative, with first coord at 0,0
    x, y = _first_move(path)
    path.move(-x, -y, inplace=True)

    # Normlize first activity to [1 0]; eliminates rotation and uniform scaling
    _, vec_first = _first_significant(_vectors(path), lambda v: v.norm(), tolerance)
    if vec_first and not vec_first.almost_equals(Vector(1, 0)):
        assert (
            vec_first.norm() > tolerance
        ), f"vec_first too close to 0-magnitude: {vec_first}"
        affinex = _affine_vec2vec(vec_first, Vector(1, 0))
        path.walk(lambda *args: _affine_callback(affinex, *args))

    # Normlize first y activity to 1.0; eliminates mirroring and non-uniform scaling
    _, vecy = _first_significant(_vectors(path), lambda v: v.y, tolerance)
    if vecy and not almost_equal(vecy.y, 1.0):
        assert vecy.norm() > tolerance, f"vecy too close to 0-magnitude: {vecy}"
        affine2 = Affine2D.identity().scale(1, 1 / vecy.y)
        path.walk(lambda *args: _affine_callback(affine2, *args))

    # TODO: what if shapes are the same but different, or different ordering, drawing cmds
    # This DOES happen in Noto; extent unclear

    path.round_multiple(tolerance, inplace=True)
    return path


def _apply_affine(affine: Affine2D, s: SVGPath) -> SVGPath:
    s_prime = copy.deepcopy(s)
    s_prime.walk(lambda *args: _affine_callback(affine, *args))
    return s_prime


def _try_affine(affine: Affine2D, s1: SVGPath, s2: SVGPath, tolerance: float):
    s1_prime = _apply_affine(affine, s1)
    return s1_prime.almost_equals(s2, tolerance)


def _round(affine, s1, s2, tolerance):
    # TODO bsearch?
    for i in _ROUND_RANGE:
        rounded = affine.round(i)
        if _try_affine(rounded, s1, s2, tolerance):
            return rounded
    return affine  # give up


def affine_between(s1: SVGShape, s2: SVGShape, tolerance: float) -> Optional[Affine2D]:
    """Returns the Affine2D to change s1 into s2 or None if no solution was found.

    Intended use is to call this only when the normalized versions of the shapes
    are the same, in which case finding a solution is typical.


    See reuse_example.html in root of picosvg for a visual explanation.

    """
    s1 = dataclasses.replace(s1, id="")
    s2 = dataclasses.replace(s2, id="")

    if s1.almost_equals(s2, tolerance):
        return Affine2D.identity()

    s1 = _affine_friendly(s1)
    s2 = _affine_friendly(s2)

    s1x, s1y = _first_move(s1)
    s2x, s2y = _first_move(s2)

    affine = Affine2D.identity().translate(s2x - s1x, s2y - s1y)
    if _try_affine(affine, s1, s2, tolerance):
        return _round(affine, s1, s2, tolerance)

    # Align the first edge with a significant x part.
    # Fixes rotation, x-scale, and uniform scaling.
    s2_vec1x_idx, s2_vec1x = _first_significant(_vectors(s2), lambda v: v.x, tolerance)
    if s2_vec1x_idx == -1:
        # bail out if we find no first edge with significant x part
        # https://github.com/googlefonts/picosvg/issues/246
        return None

    s1_vec1 = _nth_vector(s1, s2_vec1x_idx)

    s1_to_origin = Affine2D.identity().translate(-s1x, -s1y)
    s2_to_origin = Affine2D.identity().translate(-s2x, -s2y)
    s1_vec1_to_s2_vec1x = _affine_vec2vec(s1_vec1, s2_vec1x)

    # Move to s2 start
    origin_to_s2 = Affine2D.identity().translate(s2x, s2y)

    affine = Affine2D.compose_ltr((s1_to_origin, s1_vec1_to_s2_vec1x, origin_to_s2))
    if _try_affine(affine, s1, s2, tolerance):
        return _round(affine, s1, s2, tolerance)

    # Could be non-uniform scaling and/or mirroring
    # Make the aligned edge the x axis then align the first edge with a significant y part.

    # Rotate first edge to lie on x axis
    s2_vec1_angle = _angle(s2_vec1x)
    rotate_s2vec1_onto_x = Affine2D.identity().rotate(-s2_vec1_angle)
    rotate_s2vec1_off_x = Affine2D.identity().rotate(s2_vec1_angle)

    affine = Affine2D.compose_ltr(
        (s1_to_origin, s1_vec1_to_s2_vec1x, rotate_s2vec1_onto_x)
    )
    s1_prime = _apply_affine(affine, s1)

    affine = Affine2D.compose_ltr((s2_to_origin, rotate_s2vec1_onto_x))
    s2_prime = _apply_affine(affine, s2)

    # The first vector we aligned now lies on the x axis
    # Find and align the first vector that heads off into y for both
    idx, s1_vecy, s2_vecy = _first_significant_for_both(
        s1_prime, s2_prime, lambda v: v.y, tolerance
    )
    if idx != -1:
        affine = Affine2D.compose_ltr(
            (
                s1_to_origin,
                s1_vec1_to_s2_vec1x,
                # lie vec1 along x axis
                rotate_s2vec1_onto_x,
                # scale first y-vectors to match; x-parts should already match
                Affine2D.identity().scale(1.0, s2_vecy.y / s1_vecy.y),
                # restore the rotation we removed
                rotate_s2vec1_off_x,
                # drop into final position
                origin_to_s2,
            )
        )
        if _try_affine(affine, s1, s2, tolerance):
            return _round(affine, s1, s2, tolerance)

    # If we still aren't the same give up
    return None



# Copyright 2020 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict, deque
import copy
import dataclasses
from functools import lru_cache, reduce
import itertools
from lxml import etree  # pytype: disable=import-error
import re
from typing import (
    Any,
    Generator,
    Iterable,
    List,
    Mapping,
    MutableMapping,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
)
from picosvg.svg_meta import (
    ATTRIB_DEFAULTS,
    attrib_default,
    number_or_percentage,
    ntos,
    splitns,
    strip_ns,
    svgns,
    xlinkns,
    parse_css_declarations,
    parse_view_box,
    _LinkedDefault,
)
from picosvg.svg_types import *
from picosvg.svg_transform import Affine2D
import numbers

_SHAPE_CLASSES = {
    "circle": SVGCircle,
    "ellipse": SVGEllipse,
    "line": SVGLine,
    "path": SVGPath,
    "polygon": SVGPolygon,
    "polyline": SVGPolyline,
    "rect": SVGRect,
}
_GRADIENT_CLASSES = {
    "linearGradient": SVGLinearGradient,
    "radialGradient": SVGRadialGradient,
}
_CLASS_ELEMENTS = {
    v: f"{{{svgns()}}}{k}" for k, v in {**_SHAPE_CLASSES, **_GRADIENT_CLASSES}.items()
}
_SHAPE_CLASSES.update({f"{{{svgns()}}}{k}": v for k, v in _SHAPE_CLASSES.items()})

_SHAPE_FIELDS = {
    tag: tuple(f.name for f in dataclasses.fields(klass))
    for tag, klass in _SHAPE_CLASSES.items()
}
_GRADIENT_FIELDS = {
    tag: tuple(f.name for f in dataclasses.fields(klass))
    for tag, klass in _GRADIENT_CLASSES.items()
}
# Manually add stop, we don't type it
_GRADIENT_FIELDS["stop"] = tuple({"offset", "stop_color", "stop_opacity"})

_GRADIENT_COORDS = {
    "linearGradient": (("x1", "y1"), ("x2", "y2")),
    "radialGradient": (("cx", "cy"), ("fx", "fy")),
}

_VALID_FIELDS = {}
_VALID_FIELDS.update(_SHAPE_FIELDS)
_VALID_FIELDS.update(_GRADIENT_FIELDS)

_XLINK_TEMP = "xlink_"


_ATTRIB_W_CUSTOM_INHERITANCE = frozenset({"clip-path", "opacity", "transform"})


# How much error, as pct of viewbox max(w,h), is allowed on lossy ops
# For example, for a Noto Emoji with a viewBox 0 0 128 128 permit error of 0.128
_MAX_PCT_ERROR = 0.1

# When you have no viewbox, use this. Absolute value in svg units.
_DEFAULT_DEFAULT_TOLERENCE = 0.1


# Rounding for rewritten gradient matrices
_GRADIENT_TRANSFORM_NDIGITS = 6


def _clamp(value: float, minv: float = 0.0, maxv: float = 1.0) -> float:
    return max(min(value, maxv), minv)


def _xlink_href_attr_name() -> str:
    return f"{{{xlinkns()}}}href"


def _copy_new_nsmap(tree, nsm):
    new_tree = etree.Element(tree.tag, nsmap=nsm)
    new_tree.attrib.update(tree.attrib)
    new_tree[:] = tree[:]
    return new_tree


def _fix_xlink_ns(tree):
    """Fix xlink namespace problems.

    If there are xlink temps, add namespace and fix temps.
    If we declare xlink but don't use it then remove it.
    """
    xlink_nsmap = {"xlink": xlinkns()}
    if "xlink" in tree.nsmap and not len(
        tree.xpath("//*[@xlink:href]", namespaces=xlink_nsmap)
    ):
        # no reason to keep xlink
        nsm = copy.copy(tree.nsmap)
        del nsm["xlink"]
        tree = _copy_new_nsmap(tree, nsm)

    elif "xlink" not in tree.nsmap and len(tree.xpath(f"//*[@{_XLINK_TEMP}]")):
        # declare xlink and fix temps
        nsm = copy.copy(tree.nsmap)
        nsm["xlink"] = xlinkns()
        tree = _copy_new_nsmap(tree, nsm)
        for el in tree.xpath(f"//*[@{_XLINK_TEMP}]"):
            # try to retain attrib order, unexpected when they shuffle
            attrs = [(k, v) for k, v in el.attrib.items()]
            el.attrib.clear()
            for name, value in attrs:
                if name == _XLINK_TEMP:
                    name = _xlink_href_attr_name()
                el.attrib[name] = value

    return tree


def _del_attrs(el, *attr_names):
    for name in attr_names:
        if name in el.attrib:
            del el.attrib[name]


def _attr_name(field_name: str) -> str:
    return field_name.replace("_", "-")


def _field_name(attr_name: str) -> str:
    return attr_name.replace("-", "_")


def _is_defs(tag):
    return strip_ns(tag) == "defs"


def _is_shape(tag):
    return strip_ns(tag) in _SHAPE_CLASSES


def _is_gradient(tag):
    return strip_ns(tag) in _GRADIENT_CLASSES


def _is_group(tag):
    return strip_ns(tag) == "g"


def _opacity(el: etree.Element) -> float:
    return _clamp(float(el.attrib.get("opacity", 1.0)))


def _is_removable_group(el):
    """
    Groups with:

        0 < opacity < 1
        >1 child

    must be retained.

    This over-retains groups; no difference unless children overlap
    """
    if not _is_group(el):
        return False
    # no attributes makes a group meaningless
    if len(el.attrib) == 0:
        return True
    num_children = sum(1 for e in el if e.tag is not etree.Comment)

    return num_children <= 1 or _opacity(el) in {0.0, 1.0}


def _try_remove_group(group_el, push_opacity=True):
    """
    Transfer children of group to their parent if possible.

    Only groups with 0 < opacity < 1 *and* multiple children must be retained.

    This over-retains groups; no difference unless children overlap
    """
    assert _is_group(group_el)

    remove = _is_removable_group(group_el)
    opacity = _opacity(group_el)
    if remove:
        children = list(group_el)
        if group_el.getparent() is not None:
            _replace_el(group_el, list(group_el))
        if push_opacity:
            for child in children:
                if child.tag is etree.Comment:
                    continue
                _inherit_attrib({"opacity": opacity}, child)
    else:
        # We're keeping the group, but we promised groups only have opacity
        group_el.attrib.clear()
        group_el.attrib["opacity"] = ntos(opacity)
        _drop_default_attrib(group_el.attrib)
    return remove


def _element_transform(el, current_transform=Affine2D.identity()):
    attr_name = "transform"
    if _is_gradient(el.tag):
        attr_name = "gradientTransform"

    raw = el.attrib.get(attr_name, None)
    if raw:
        return Affine2D.compose_ltr((Affine2D.fromstring(raw), current_transform))
    return current_transform


def from_element(el, **inherited_attrib):
    if not _is_shape(el.tag):
        raise ValueError(f"Bad tag <{el.tag}>")
    data_type = _SHAPE_CLASSES[el.tag]
    attrs = {**inherited_attrib, **el.attrib}
    args = {
        f.name: f.type(attrs[_attr_name(f.name)])
        for f in dataclasses.fields(data_type)
        if attrs.get(_attr_name(f.name), "").strip()
    }
    return data_type(**args)


def to_element(data_obj, **inherited_attrib):
    el = etree.Element(_CLASS_ELEMENTS[type(data_obj)])
    for field in dataclasses.fields(data_obj):
        attr_name = _attr_name(field.name)
        field_value = getattr(data_obj, field.name)
        # omit attributes whose value == the respective default,
        # unless it's != from the attribute value inherited from context
        if isinstance(field.default, _LinkedDefault):
            default_value = field.default(data_obj)
        else:
            default_value = field.default
        attrib_value = field_value
        if isinstance(attrib_value, numbers.Number):
            attrib_value = ntos(attrib_value)
        elif isinstance(attrib_value, Affine2D):
            attrib_value = attrib_value.tostring()
        if attr_name in inherited_attrib:
            if attrib_value == inherited_attrib[attr_name]:
                continue
        elif field_value == default_value:
            continue
        el.attrib[attr_name] = attrib_value
    return el


def _reset_attrs(data_obj, field_pred):
    for field in dataclasses.fields(data_obj):
        if field_pred(field):
            setattr(data_obj, field.name, field.default)


def _safe_remove(el: etree.Element):
    parent = el.getparent()
    if parent is not None:
        parent.remove(el)


def _id_of_target(url):
    match = re.match(r"^url[(]#([\w-]+)[)]$", url)
    if not match:
        raise ValueError(f'Unrecognized url "{url}"')
    return match.group(1)


def _xpath_for_url(url, el_tag):
    return f'//svg:{el_tag}[@id="{_id_of_target(url)}"]'


def _attrib_to_pass_on(current_attrib, el, skips=_ATTRIB_W_CUSTOM_INHERITANCE):
    attr_catcher = etree.Element("dummy")
    _inherit_attrib(el.attrib, attr_catcher, skips=skips, skip_unhandled=True)
    _inherit_attrib(current_attrib, attr_catcher, skips=skips)
    return dict(attr_catcher.attrib)


def _replace_el(el, replacements):
    parent = el.getparent()
    idx = parent.index(el)
    parent.remove(el)
    for child_idx, child in enumerate(replacements):
        parent.insert(idx + child_idx, child)


class SVGTraverseContext(NamedTuple):
    nth_of_type: int
    element: etree.Element
    path: str
    transform: Affine2D
    clips: Tuple[SVGPath, ...]
    attrib: Mapping[str, Any]  # except _ATTRIB_W_CUSTOM_INHERITANCE

    def depth(self) -> int:
        return self.path.count("/") - 1

    def shape(self) -> SVGShape:
        return from_element(self.element, **self.attrib)

    def is_shape(self):
        return _is_shape(self.element)

    def is_group(self):
        return _is_group(self.element)


class SVG:

    svg_root: etree.Element
    elements: List[Tuple[etree.Element, Tuple[SVGShape, ...]]]

    def __init__(self, svg_root):
        self.svg_root = svg_root
        self.elements = []

    def _elements(self) -> List[Tuple[etree.Element, Tuple[SVGShape, ...]]]:
        if self.elements:
            return self.elements
        elements = []
        for context in self.depth_first(resolve_clip_paths=False):
            el = context.element
            if el.tag not in _SHAPE_CLASSES:
                continue
            elements.append((el, (context.shape(),)))
        self.elements = elements
        return self.elements

    def _set_element(self, idx: int, el: etree.Element, shapes: Tuple[SVGShape, ...]):
        self.elements[idx] = (el, shapes)

    def view_box(self) -> Optional[Rect]:
        if "viewBox" not in self.svg_root.attrib:
            # if there is no explicit viewbox try to use width/height
            w = self.svg_root.attrib.get("width", None)
            h = self.svg_root.attrib.get("height", None)
            if w and h:
                return Rect(0, 0, float(w), float(h))
            else:
                return None

        return parse_view_box(self.svg_root.attrib["viewBox"])

    def _default_tolerance(self):
        vbox = self.view_box()
        # Absence of viewBox is unusual
        if vbox is None:
            return _DEFAULT_DEFAULT_TOLERENCE
        return min(vbox.w, vbox.h) * _MAX_PCT_ERROR / 100

    @property
    def tolerance(self):
        return self._default_tolerance()

    def shapes(self):
        """Returns all shapes in order encountered.

        Use to operate per-shape; if you want to iterate over graph use breadth_first.
        """
        return tuple(shape for (_, shapes) in self._elements() for shape in shapes)

    def absolute(self, inplace=False):
        """Converts all basic shapes to their equivalent path."""
        if not inplace:
            svg = SVG(copy.deepcopy(self.svg_root))
            svg.absolute(inplace=True)
            return svg

        swaps = []
        for idx, (el, (shape,)) in enumerate(self._elements()):
            self.elements[idx] = (el, (shape.absolute(),))
        return self

    def shapes_to_paths(self, inplace=False):
        """Converts all basic shapes to their equivalent path."""
        if not inplace:
            svg = SVG(copy.deepcopy(self.svg_root))
            svg.shapes_to_paths(inplace=True)
            return svg

        swaps = []
        for idx, (el, (shape,)) in enumerate(self._elements()):
            self.elements[idx] = (el, (shape.as_path(),))
        return self

    def expand_shorthand(self, inplace=False):
        if not inplace:
            svg = SVG(copy.deepcopy(self.svg_root))
            svg.expand_shorthand(inplace=True)
            return svg

        for idx, (el, (shape,)) in enumerate(self._elements()):
            if isinstance(shape, SVGPath):
                self.elements[idx] = (
                    el,
                    (shape.explicit_lines().expand_shorthand(inplace=True),),
                )
        return self

    def _apply_styles(self, el: etree.Element):
        parse_css_declarations(el.attrib.pop("style", ""), el.attrib)

    def apply_style_attributes(self, inplace=False):
        """Converts inlined CSS "style" attributes to equivalent SVG attributes."""
        if not inplace:
            svg = SVG(copy.deepcopy(self.svg_root))
            svg.apply_style_attributes(inplace=True)
            return svg

        if self.elements:
            # if we already parsed the SVG shapes, apply style attrs and sync tree
            for shape in self.shapes():
                shape.apply_style_attribute(inplace=True)
            self._update_etree()

        # parse all remaining style attributes (e.g. in gradients or root svg element)
        for el in itertools.chain((self.svg_root,), self.xpath("//svg:*[@style]")):
            self._apply_styles(el)

        return self

    def xpath(self, xpath: str, el: etree.Element = None, expected_result_range=None):
        if el is None:
            el = self.svg_root
        results = el.xpath(xpath, namespaces={"svg": svgns()})
        if expected_result_range and len(results) not in expected_result_range:
            raise ValueError(
                f"Expected {xpath} matches in {expected_result_range}, {len(results)} results"
            )
        return results

    def xpath_one(self, xpath):
        return self.xpath(xpath, expected_result_range=range(1, 2))[0]

    def resolve_url(self, url, el_tag):
        return self.xpath_one(_xpath_for_url(url, el_tag))

    def _resolve_use(self, scope_el):
        attrib_not_copied = {
            "x",
            "y",
            "width",
            "height",
            "transform",
            _xlink_href_attr_name(),
        }

        # capture elements by id so even if we change it they remain stable
        el_by_id = {el.attrib["id"]: el for el in self.xpath(".//svg:*[@id]")}

        while True:
            swaps = []
            use_els = list(self.xpath(".//svg:use", el=scope_el))
            if not use_els:
                break
            for use_el in use_els:
                ref = use_el.attrib.get(_xlink_href_attr_name(), "")
                if not ref.startswith("#"):
                    raise ValueError(f"Only use #fragment supported, reject {ref}")

                target = el_by_id.get(ref[1:], None)
                if target is None:
                    raise ValueError(f"No element has id '{ref[1:]}'")

                new_el = copy.deepcopy(target)
                # leaving id's on <use> instantiated content is a path to duplicate ids
                for el in new_el.getiterator("*"):
                    if "id" in el.attrib:
                        del el.attrib["id"]

                group = etree.Element(f"{{{svgns()}}}g", nsmap=self.svg_root.nsmap)
                affine = Affine2D.identity().translate(
                    float(use_el.attrib.get("x", 0)), float(use_el.attrib.get("y", 0))
                )

                if "transform" in use_el.attrib:
                    affine = Affine2D.compose_ltr(
                        (
                            affine,
                            Affine2D.fromstring(use_el.attrib["transform"]),
                        )
                    )

                if affine != Affine2D.identity():
                    group.attrib["transform"] = affine.tostring()

                for attr_name in use_el.attrib:
                    if attr_name in attrib_not_copied:
                        continue
                    group.attrib[attr_name] = use_el.attrib[attr_name]

                group.append(new_el)

                if _try_remove_group(group, push_opacity=False):
                    _inherit_attrib(group.attrib, new_el)
                    swaps.append((use_el, new_el))
                else:
                    swaps.append((use_el, group))

            for old_el, new_el in swaps:
                old_el.getparent().replace(old_el, new_el)

    def resolve_use(self, inplace=False):
        """Instantiate reused elements.

        https://www.w3.org/TR/SVG11/struct.html#UseElement"""
        if not inplace:
            svg = SVG(copy.deepcopy(self.svg_root))
            svg.resolve_use(inplace=True)
            return svg

        self._update_etree()
        self._resolve_use(self.svg_root)
        return self

    def _resolve_clip_path(
        self, clip_path_url, transform=Affine2D.identity()
    ) -> SVGPath:
        clip_path_el = self.resolve_url(clip_path_url, "clipPath")
        self._resolve_use(clip_path_el)

        transform = _element_transform(clip_path_el, transform)
        clip_paths = [
            from_element(e).apply_transform(_element_transform(e, transform))
            for e in clip_path_el
        ]

        clip = SVGPath.from_commands(union(clip_paths))

        if "clip-path" in clip_path_el.attrib:
            # TODO cycle detection
            clip_clop = self._resolve_clip_path(
                clip_path_el.attrib["clip-path"], transform
            )
            clip = SVGPath.from_commands(intersection([clip, clip_clop]))

        return clip

    def append_to(self, xpath, el):
        self._update_etree()
        self.xpath_one(xpath).append(el)
        return el

    def _new_id(self, template):
        for i in range(1 << 16):
            potential_id = template % i
            existing = self.xpath(f'//svg:*[@id="{potential_id}"]')
            if not existing:
                return potential_id
        raise ValueError(f"No free id for {template}")

    def _traverse(self, next_fn, append_fn, resolve_clip_paths=True):
        frontier = [
            SVGTraverseContext(
                0,
                self.svg_root,
                "/svg[0]",
                Affine2D.identity(),
                (),
                _attrib_to_pass_on(_INHERITABLE_ATTRIB_DEFAULTS, self.svg_root),
            )
        ]
        while frontier:
            context = next_fn(frontier)
            yield context

            child_idxs = defaultdict(int)
            new_entries = []
            for child in context.element:
                if child.tag is etree.Comment:
                    continue
                transform = _element_transform(child, context.transform)
                clips = context.clips
                if (
                    resolve_clip_paths
                    and child.attrib.get("clip-path")  # neither None nor ""
                    and child.attrib["clip-path"] != "none"
                ):
                    clips += (
                        self._resolve_clip_path(child.attrib["clip-path"], transform),
                    )

                nth_of_type = child_idxs[strip_ns(child.tag)]
                child_idxs[strip_ns(child.tag)] += 1
                path = f"{context.path}/{strip_ns(child.tag)}[{nth_of_type}]"
                child_context = SVGTraverseContext(
                    nth_of_type,
                    child,
                    path,
                    transform,
                    clips,
                    _attrib_to_pass_on(context.attrib, child),
                )
                new_entries.append(child_context)
            append_fn(frontier, new_entries)

    def depth_first(self, resolve_clip_paths=True):
        # dfs will take from the back
        # reverse so this still yields in order (first child, second child, etc)
        # makes processing feel more intuitive
        yield from self._traverse(
            lambda f: f.pop(),
            lambda f, e: f.extend(reversed(e)),
            resolve_clip_paths=resolve_clip_paths,
        )

    def breadth_first(self, resolve_clip_paths=True):
        yield from self._traverse(
            lambda f: f.pop(0),
            lambda f, e: f.extend(e),
            resolve_clip_paths=resolve_clip_paths,
        )

    def _add_to_defs(self, defs, new_el):
        if "id" not in new_el.attrib:
            return  # idless defs are useless
        new_id = new_el.attrib["id"]
        insert_at = 0
        for i, el in enumerate(defs):
            if new_id < el.attrib["id"]:
                insert_at = i
                break
        defs.insert(insert_at, new_el)

    def _transformed_gradient(self, defs, fill_el, transform, shape_bbox):
        assert _is_gradient(fill_el), f"Not sure how to fill from {fill_el.tag}"

        gradient = (
            _GRADIENT_CLASSES[strip_ns(fill_el.tag)]
            .from_element(fill_el, self.view_box())
            .as_user_space_units(shape_bbox, inplace=True)
        )
        gradient.gradientTransform = Affine2D.compose_ltr(
            (gradient.gradientTransform, transform)
        ).round(_GRADIENT_TRANSFORM_NDIGITS)
        gradient.id = self._new_id(gradient.id + "_%d")

        new_fill = to_element(gradient)
        # TODO normalize stop elements too
        new_fill.extend(copy.deepcopy(stop) for stop in fill_el)

        self._apply_gradient_translation(new_fill)

        self._add_to_defs(defs, new_fill)
        return new_fill

    def _simplify(self):
        """
        Removes groups where possible, applies transforms, applies clip paths.
        """
        # Reversed: we want leaves first
        to_process = reversed(tuple(c for c in self.breadth_first()))

        defs = etree.Element(f"{{{svgns()}}}defs", nsmap=self.svg_root.nsmap)
        self.svg_root.insert(0, defs)

        for context in to_process:
            if "clipPath" in context.path:
                _safe_remove(context.element)
                continue

            el = context.element
            _del_attrs(el, "clip-path", "transform")  # handled separately
            _inherit_attrib(context.attrib, el)

            # Only some elements change
            if _is_shape(el.tag):
                assert len(el) == 0, "Shapes shouldn't have children"

                # If we are transformed and we use a gradient we may need to
                # emit the transformed gradient
                if context.transform != Affine2D.identity() and "url" in el.attrib.get(
                    "fill", ""
                ):
                    fill_el = self.resolve_url(el.attrib["fill"], "*")
                    self._apply_gradient_template(fill_el)
                    fill_el = self._transformed_gradient(
                        defs,
                        fill_el,
                        context.transform,
                        from_element(el).bounding_box(),
                    )
                    fill_id = fill_el.attrib["id"]
                    el.attrib["fill"] = f"url(#{fill_id})"

                paths = [from_element(el).as_path().absolute(inplace=True)]
                initial_path = copy.deepcopy(paths[0])

                # stroke may introduce multiple paths
                assert len(paths) == 1  # oh ye of little faith
                if paths[0].stroke != "none":
                    paths = list(self._stroke(paths[0]))

                # Any remaining stroke attributes don't do anything
                # For example, a stroke-width with no stroke set is removed
                for path in paths:
                    _reset_attrs(path, lambda field: field.name.startswith("stroke"))

                # Apply any transform
                if context.transform != Affine2D.identity():
                    paths = [p.apply_transform(context.transform) for p in paths]

                if context.clips:
                    for p in paths:
                        # When constructing pathops.Path objects for performing the
                        # intersection operation, we need to use the fill-rule attribute
                        # for the shape to be clipped, and clip-rule for the clipping
                        # path itself (clip-rule only applies within clipPath element).
                        p.update_path(
                            intersection(
                                (p, *context.clips),
                                fill_rules=(
                                    p.fill_rule,
                                    *(c.clip_rule for c in context.clips),
                                ),
                            ),
                            inplace=True,
                        )
                        # skia-pathops operations always return nonzero winding paths
                        p.fill_rule = "nonzero"

                if len(paths) != 1 or paths[0] != initial_path:
                    _replace_el(el, [to_element(p) for p in paths])

            elif _is_gradient(el.tag):
                _safe_remove(el)
                self._add_to_defs(defs, el)
                self._apply_gradient_template(el)
                self._apply_gradient_translation(el)

            elif _is_defs(el.tag):
                # any children were already processed
                # now just moved to master defs
                for child_el in el:
                    self._add_to_defs(defs, child_el)
                _safe_remove(el)

            elif _is_group(el.tag):
                _try_remove_group(el)

        # https://github.com/googlefonts/nanoemoji/issues/275
        _del_attrs(self.svg_root, *_INHERITABLE_ATTRIB)

        self._remove_orphaned_gradients()

        # After simplification only gradient defs should be referenced
        # It's illegal for picosvg to leave anything else in defs
        for unused_el in [el for el in defs if not _is_gradient(el)]:
            defs.remove(unused_el)

        self.elements = None  # force elements to reload

    def simplify(self, inplace=False):
        if not inplace:
            svg = SVG(copy.deepcopy(self.svg_root))
            svg.simplify(inplace=True)
            return svg

        self._update_etree()
        self._simplify()
        return self

    def _stroke(self, shape):
        """Convert stroke to path.

        Returns sequence of shapes in draw order. That is, result[1] should be
        drawn on top of result[0], etc."""

        assert shape.stroke != "none"

        # make a new path that is the stroke
        stroke = shape.as_path().update_path(shape.stroke_commands(self.tolerance))

        # skia stroker returns paths with 'nonzero' winding fill rule
        stroke.fill_rule = stroke.clip_rule = "nonzero"

        # a few attributes move in interesting ways
        stroke.opacity *= stroke.stroke_opacity
        stroke.fill = stroke.stroke
        # the fill and stroke are now different (filled) paths, reset 'fill_opacity'
        # to default and only use a combined 'opacity' in each one.
        shape.opacity *= shape.fill_opacity
        shape.fill_opacity = stroke.fill_opacity = 1.0

        # remove all the stroke settings
        for cleanmeup in (shape, stroke):
            _reset_attrs(cleanmeup, lambda field: field.name.startswith("stroke"))

        if not shape.might_paint():
            return (stroke,)

        # The original id doesn't correctly refer to either
        # It would be for the best if any id-based operations happened first
        shape.id = stroke.id = ""

        return (shape, stroke)

    def clip_to_viewbox(self, inplace=False):
        if not inplace:
            svg = SVG(copy.deepcopy(self.svg_root))
            svg.clip_to_viewbox(inplace=True)
            return svg

        self._update_etree()

        view_box = self.view_box()

        # phase 1: dump shapes that are completely out of bounds
        for el, (shape,) in self._elements():
            if view_box.intersection(shape.bounding_box()) is None:
                _safe_remove(el)

        self.elements = None  # force elements to reload

        # phase 2: clip things that are partially out of bounds
        updates = []
        for idx, (el, (shape,)) in enumerate(self._elements()):
            bbox = shape.bounding_box()
            isct = view_box.intersection(bbox)
            assert isct is not None, f"We should have already dumped {shape}"
            if bbox == isct:
                continue
            clip_path = (
                SVGRect(x=isct.x, y=isct.y, width=isct.w, height=isct.h)
                .as_path()
                .absolute(inplace=True)
            )
            shape = shape.as_path().absolute(inplace=True)
            shape.update_path(
                intersection(
                    (shape, clip_path),
                    fill_rules=(shape.fill_rule, clip_path.clip_rule),
                ),
                inplace=True,
            )
            shape.fill_rule = "nonzero"
            updates.append((idx, el, shape))

        for idx, el, shape in updates:
            self._set_element(idx, el, (shape,))

        # Update the etree
        self._update_etree()

        # We may now have useless groups
        for context in reversed(list(self.depth_first())):
            if _is_group(context.element):
                _try_remove_group(context.element)

        return self

    def evenodd_to_nonzero_winding(self, inplace=False):
        if not inplace:
            svg = SVG(copy.deepcopy(self.svg_root))
            svg.evenodd_to_nonzero_winding(inplace=True)
            return svg

        for idx, (el, (shape,)) in enumerate(self._elements()):
            if shape.fill_rule == "evenodd":
                path = shape.as_path().remove_overlaps(inplace=True)
                self._set_element(idx, el, (path,))

        return self

    def round_floats(self, ndigits: int, inplace=False):
        if not inplace:
            svg = SVG(copy.deepcopy(self.svg_root))
            svg.round_floats(ndigits, inplace=True)
            return svg

        for shape in self.shapes():
            shape.round_floats(ndigits, inplace=True)
        return self

    def remove_unpainted_shapes(self, inplace=False):
        if not inplace:
            svg = SVG(copy.deepcopy(self.svg_root))
            svg.remove_unpainted_shapes(inplace=True)
            return svg

        self._update_etree()

        remove = []
        for (el, (shape,)) in self._elements():
            if not shape.might_paint():
                remove.append(el)

        for el in remove:
            el.getparent().remove(el)

        self.elements = None

        return self

    def remove_nonsvg_content(self, inplace=False):
        if not inplace:
            svg = SVG(copy.deepcopy(self.svg_root))
            svg.remove_nonsvg_content(inplace=True)
            return svg

        self._update_etree()

        good_ns = {svgns(), xlinkns()}
        if self.svg_root.nsmap[None] == svgns():
            good_ns.add(None)

        el_to_rm = []
        for el in self.svg_root.getiterator("*"):
            attr_to_rm = []
            ns, _ = splitns(el.tag)
            if ns not in good_ns:
                el_to_rm.append(el)
                continue
            for attr in el.attrib:
                ns, _ = splitns(attr)
                if ns not in good_ns:
                    attr_to_rm.append(attr)
            for attr in attr_to_rm:
                del el.attrib[attr]

        for el in el_to_rm:
            el.getparent().remove(el)

        # Make svg default; destroy anything unexpected
        good_nsmap = {
            None: svgns(),
            "xlink": xlinkns(),
        }
        if any(good_nsmap.get(k, None) != v for k, v in self.svg_root.nsmap.items()):
            self.svg_root = _copy_new_nsmap(self.svg_root, good_nsmap)

        self.elements = None

        return self

    def remove_comments(self, inplace=False):
        if not inplace:
            svg = SVG(copy.deepcopy(self.svg_root))
            svg.remove_comments(inplace=True)
            return svg

        self._update_etree()

        for el in self.xpath("//comment()"):
            el.getparent().remove(el)

        return self

    def remove_anonymous_symbols(self, inplace=False):
        # No id makes a symbol useless
        # https://github.com/googlefonts/picosvg/issues/46
        if not inplace:
            svg = SVG(copy.deepcopy(self.svg_root))
            svg.remove_anonymous_symbols(inplace=True)
            return svg

        self._update_etree()

        for el in self.xpath("//svg:symbol[not(@id)]"):
            el.getparent().remove(el)

        return self

    def remove_title_meta_desc(self, inplace=False):
        if not inplace:
            svg = SVG(copy.deepcopy(self.svg_root))
            svg.remove_title_meta_desc(inplace=True)
            return svg

        self._update_etree()

        for tag in ("title", "desc", "metadata", "comment"):
            for el in self.xpath(f"//svg:{tag}"):
                el.getparent().remove(el)

        return self

    def set_attributes(self, name_values, xpath="/svg:svg", inplace=False):
        if not inplace:
            svg = SVG(copy.deepcopy(self.svg_root))
            svg.set_attributes(name_values, xpath=xpath, inplace=True)
            return svg

        self._update_etree()

        for el in self.xpath(xpath):
            for name, value in name_values:
                el.attrib[name] = value

        return self

    def remove_attributes(self, names, xpath="/svg:svg", inplace=False):
        """Drop things like viewBox, width, height that set size of overall svg"""
        if not inplace:
            svg = SVG(copy.deepcopy(self.svg_root))
            svg.remove_attributes(names, xpath=xpath, inplace=True)
            return svg

        self._update_etree()

        for el in self.xpath(xpath):
            _del_attrs(el, *names)

        return self

    def normalize_opacity(self, inplace=False):
        """Merge '{fill,stroke}_opacity' with generic 'opacity' when possible."""
        if not inplace:
            svg = SVG(copy.deepcopy(self.svg_root))
            svg.normalize_opacity(inplace=True)
            return svg

        for shape in self.shapes():
            shape.normalize_opacity(inplace=True)

        return self

    def _iter_nested_svgs(
        self, root: etree.Element
    ) -> Generator[etree.Element, None, None]:
        # This is different from Element.iter("svg") in that we don't yield the root
        # svg element itself, only traverse its children and yield any immediate
        # nested SVGs without traversing the latter's children as well.
        frontier = deque(root)
        while frontier:
            el = frontier.popleft()
            if el.tag is etree.Comment:
                continue
            if strip_ns(el.tag) == "svg":
                yield el
            elif len(el) != 0:
                frontier.extend(el)

    def _unnest_svg(
        self, svg: etree.Element, parent_width: float, parent_height: float
    ) -> Tuple[etree.Element, ...]:
        x = float(svg.attrib.get("x", 0))
        y = float(svg.attrib.get("y", 0))
        width = float(svg.attrib.get("width", parent_width))
        height = float(svg.attrib.get("height", parent_height))

        viewport = viewbox = Rect(x, y, width, height)
        if "viewBox" in svg.attrib:
            viewbox = parse_view_box(svg.attrib["viewBox"])

        # first recurse to un-nest any nested nested SVGs
        self._swap_elements(
            (el, self._unnest_svg(el, viewbox.w, viewbox.h))
            for el in self._iter_nested_svgs(svg)
        )

        g = etree.Element(f"{{{svgns()}}}g")
        g.extend(svg)

        if viewport != viewbox:
            preserve_aspect_ratio = svg.attrib.get("preserveAspectRatio", "xMidYMid")
            transform = Affine2D.rect_to_rect(viewbox, viewport, preserve_aspect_ratio)
        else:
            transform = Affine2D.identity().translate(x, y)

        if "transform" in svg.attrib:
            transform = Affine2D.compose_ltr(
                (transform, Affine2D.fromstring(svg.attrib["transform"]))
            )

        if transform != Affine2D.identity():
            g.attrib["transform"] = transform.tostring()

        # non-root svg elements by default have overflow="hidden" which means a clip path
        # the size of the SVG viewport is applied; if overflow="visible" don't clip
        # https://www.w3.org/TR/SVG/render.html#OverflowAndClipProperties
        overflow = svg.attrib.get("overflow", "hidden")
        if overflow == "visible":
            return (g,)

        if overflow != "hidden":
            raise NotImplementedError(f"overflow='{overflow}' is not supported")

        clip_path = etree.Element(
            f"{{{svgns()}}}clipPath", {"id": self._new_id("nested-svg-viewport-%d")}
        )
        clip_path.append(to_element(SVGRect(x=x, y=y, width=width, height=height)))
        clipped_g = etree.Element(f"{{{svgns()}}}g")
        clipped_g.attrib["clip-path"] = f"url(#{clip_path.attrib['id']})"
        clipped_g.append(g)

        return (clip_path, clipped_g)

    def resolve_nested_svgs(self, inplace=False):
        """Replace nested <svg> elements with equivalent <g> with a transform.

        NOTE: currently this is still missing two features:
        1) resolving percentage units in reference to the nearest SVG viewport;
        2) applying a clip to all children of the nested SVG with a rectangle the size
           of the new viewport (inner SVGs have default overflow property set to
           'hidden'). Blocked on https://github.com/googlefonts/picosvg/issues/200
        No error is raised in these cases.

        References:
        - https://www.w3.org/TR/SVG/coords.html
        - https://www.sarasoueidan.com/blog/nesting-svgs/
        """
        if not inplace:
            svg = SVG(copy.deepcopy(self.svg_root))
            svg.resolve_nested_svgs(inplace=True)
            return svg

        self._update_etree()

        nested_svgs = list(self._iter_nested_svgs(self.svg_root))
        if len(nested_svgs) == 0:
            return

        vb = self.view_box()
        if vb is None:
            raise ValueError(
                "Can't determine root SVG width/height, "
                "which is required for resolving nested SVGs"
            )

        self._swap_elements(
            (el, self._unnest_svg(el, vb.w, vb.h)) for el in nested_svgs
        )

        return self

    def _select_gradients(self):
        return self.xpath(" | ".join(f"//svg:{tag}" for tag in _GRADIENT_CLASSES))

    def _apply_gradient_translation(self, el: etree.Element):
        assert _is_gradient(el)
        gradient = _GRADIENT_CLASSES[strip_ns(el.tag)].from_element(el, self.view_box())
        affine = gradient.gradientTransform

        # split translation from rest of the transform and apply to gradient coords
        translate, affine_prime = affine.decompose_translation()
        if translate.round(_GRADIENT_TRANSFORM_NDIGITS) != Affine2D.identity():
            for x_attr, y_attr in _GRADIENT_COORDS[strip_ns(el.tag)]:
                x = getattr(gradient, x_attr)
                y = getattr(gradient, y_attr)
                x_prime, y_prime = translate.map_point((x, y))
                setattr(gradient, x_attr, round(x_prime, _GRADIENT_TRANSFORM_NDIGITS))
                setattr(gradient, y_attr, round(y_prime, _GRADIENT_TRANSFORM_NDIGITS))

        gradient.gradientTransform = affine_prime.round(_GRADIENT_TRANSFORM_NDIGITS)

        el.attrib.clear()
        el.attrib.update(to_element(gradient).attrib)

    def _apply_gradient_template(self, gradient: etree.Element):
        # Gradients can have an 'href' attribute that specifies another gradient as
        # a template, inheriting its attributes and/or stops when not already defined:
        # https://www.w3.org/TR/SVG/pservers.html#PaintServerTemplates

        assert _is_gradient(gradient)

        href_attr = _xlink_href_attr_name()
        if href_attr not in gradient.attrib:
            return  # nop

        ref = gradient.attrib[href_attr]
        if not ref.startswith("#"):
            raise ValueError(f"Only use #fragment supported, reject {ref}")
        ref = ref[1:].strip()

        template = self.xpath_one(f'.//svg:*[@id="{ref}"]')

        template_tag = strip_ns(template.tag)
        if template_tag not in _GRADIENT_CLASSES:
            raise ValueError(
                f"Referenced element with id='{ref}' has unexpected tag: "
                f"expected linear or radialGradient, found '{template_tag}'"
            )

        # recurse if template references another template
        if template.attrib.get(href_attr):
            self._apply_gradient_template(template)

        for attr_name in _GRADIENT_FIELDS[strip_ns(gradient.tag)]:
            if attr_name in template.attrib and attr_name not in gradient.attrib:
                gradient.attrib[attr_name] = template.attrib[attr_name]

        # only copy stops if we don't have our own
        if len(gradient) == 0:
            for stop_el in template:
                new_stop_el = copy.deepcopy(stop_el)
                # strip stop id if present; useless and no longer unique
                _del_attrs(new_stop_el, "id")
                gradient.append(new_stop_el)

        del gradient.attrib[href_attr]

    def _remove_orphaned_gradients(self):
        # remove orphaned templates, only keep gradients directly referenced by shapes
        used_gradient_ids = set()
        for shape in self.shapes():
            if shape.fill.startswith("url("):
                try:
                    el = self.resolve_url(shape.fill, "*")
                except ValueError:  # skip not found
                    continue
                if strip_ns(el.tag) not in _GRADIENT_CLASSES:
                    # unlikely the url target isn't a gradient but I'm not the police
                    continue
                used_gradient_ids.add(el.attrib["id"])
        for grad in self._select_gradients():
            if grad.attrib.get("id") not in used_gradient_ids:
                _safe_remove(grad)

    def checkpicosvg(self):
        """Check for nano violations, return xpaths to bad elements.

        If result sequence empty then this is a valid picosvg.
        """

        self._update_etree()

        errors = []
        bad_paths = set()

        path_allowlist = {
            r"^/svg\[0\]$",
            r"^/svg\[0\]/defs\[0\]$",
            r"^/svg\[0\]/defs\[0\]/(linear|radial)Gradient\[\d+\](/stop\[\d+\])?$",
            r"^/svg\[0\](/(path|g)\[\d+\])+$",
        }
        paths_required = {
            "/svg[0]",
            "/svg[0]/defs[0]",
        }

        # Make a list of xpaths with offsets (/svg/defs[0]/..., etc)
        ids = {}
        for context in self.breadth_first():
            if any(context.path.startswith(bp) for bp in bad_paths):
                continue  # no sense reporting all the children as bad

            if not any((re.match(pat, context.path) for pat in path_allowlist)):
                errors.append(f"BadElement: {context.path}")
                bad_paths.add(context.path)
                continue

            paths_required.discard(context.path)

            el_id = context.element.attrib.get("id", None)
            if el_id is not None:
                if el_id in ids:
                    errors.append(
                        f'BadElement: {context.path} reuses id="{el_id}", first seen at {ids[el_id]}'
                    )
                ids[el_id] = context.path

        for path in paths_required:
            errors.append(f"MissingElement: {path}")

        # TODO paths, groups, & gradients should only have specific attributes

        return tuple(errors)

    def topicosvg(self, *, ndigits=3, inplace=False):
        if not inplace:
            svg = SVG(copy.deepcopy(self.svg_root))
            svg.topicosvg(inplace=True)
            return svg

        self._update_etree()

        # Discard useless content
        self.remove_nonsvg_content(inplace=True)
        self.remove_comments(inplace=True)
        self.remove_anonymous_symbols(inplace=True)
        self.remove_title_meta_desc(inplace=True)

        # Simplify things that simplify in isolation
        self.apply_style_attributes(inplace=True)
        self.resolve_nested_svgs(inplace=True)
        self.shapes_to_paths(inplace=True)
        self.expand_shorthand(inplace=True)
        self.resolve_use(inplace=True)

        # Simplify things that do not simplify in isolation
        self.simplify(inplace=True)

        # Tidy up
        self.evenodd_to_nonzero_winding(inplace=True)
        self.remove_unpainted_shapes(inplace=True)
        self.normalize_opacity(inplace=True)
        self.absolute(inplace=True)
        self.round_floats(ndigits, inplace=True)

        nano_violations = self.checkpicosvg()
        if nano_violations:
            raise ValueError(
                "Unable to convert to picosvg: " + ",".join(nano_violations)
            )

        return self

    @staticmethod
    def _swap_elements(swaps: Iterable[Tuple[etree.Element, Sequence[etree.Element]]]):
        for old_el, new_els in swaps:
            for new_el in reversed(new_els):
                old_el.addnext(new_el)
            parent = old_el.getparent()
            if parent is None:
                raise ValueError("Lost parent!")
            parent.remove(old_el)

    @lru_cache(maxsize=None)
    def _inherited_attrib(self, el: etree.Element) -> Mapping[str, str]:
        parents = []
        while el.getparent() is not None:
            el = el.getparent()
            parents.append(el)
        return reduce(
            _attrib_to_pass_on, reversed(parents), _INHERITABLE_ATTRIB_DEFAULTS
        )

    def _update_etree(self):
        if not self.elements:
            return
        self._inherited_attrib.cache_clear()
        self._swap_elements(
            (
                old_el,
                [to_element(s, **self._inherited_attrib(old_el)) for s in shapes],
            )
            for old_el, shapes in self.elements
        )
        self.elements = None

    def toetree(self):
        self._update_etree()
        self.svg_root = _fix_xlink_ns(self.svg_root)
        return copy.deepcopy(self.svg_root)

    def tostring(self, pretty_print=False):
        return etree.tostring(self.toetree(), pretty_print=pretty_print).decode("utf-8")

    @classmethod
    def fromstring(cls, string):
        if isinstance(string, bytes):
            string = string.decode("utf-8")

        # svgs are fond of not declaring xlink
        # based on https://mailman-mail5.webfaction.com/pipermail/lxml/20100323/021184.html
        if "xlink" in string and "xmlns:xlink" not in string:
            string = string.replace("xlink:href", _XLINK_TEMP)

        # encode because fromstring dislikes xml encoding decl if input is str
        parser = etree.XMLParser(remove_blank_text=True)
        tree = etree.fromstring(string.encode("utf-8"), parser)
        tree = _fix_xlink_ns(tree)
        return cls(tree)

    @classmethod
    def parse(cls, file_or_path):
        if hasattr(file_or_path, "read"):
            raw_svg = file_or_path.read()
        else:
            with open(file_or_path) as f:
                raw_svg = f.read()
        return cls.fromstring(raw_svg)


def _inherit_copy(attrib, child, attr_name):
    if attr_name in child.attrib:
        return
    if attr_name in attrib:
        child.attrib[attr_name] = attrib[attr_name]


def _inherit_multiply(attrib, child, attr_name):
    if attr_name not in attrib and attr_name not in child.attrib:
        return
    value = float(attrib.get(attr_name, 1.0))
    value *= float(child.attrib.get(attr_name, 1.0))
    child.attrib[attr_name] = ntos(value)


def _inherit_clip_path(attrib, child, attr_name):
    clips = sorted(
        child.attrib.get("clip-path", "").split(",") + [attrib.get("clip-path", "")]
    )
    child.attrib["clip-path"] = ",".join([c for c in clips if c])


def _inherit_nondefault_overflow(attrib, child, attr_name):
    value = attrib.get(attr_name, "visible")
    if value != "visible":
        _inherit_copy(attrib, child, attr_name)


def _inherit_matrix_multiply(attrib, child, attr_name):
    transform = Affine2D.identity()
    if attr_name in attrib:
        transform = Affine2D.fromstring(attrib[attr_name])
    if attr_name in child.attrib:
        transform = Affine2D.compose_ltr(
            (Affine2D.fromstring(child.attrib[attr_name]), transform)
        )
    if transform != Affine2D.identity():
        child.attrib[attr_name] = transform.tostring()
    else:
        del child.attrib[attr_name]


def _do_not_inherit(*_):
    return


_INHERIT_ATTRIB_HANDLERS = {
    "clip-rule": _inherit_copy,
    "color": _inherit_copy,
    "display": _inherit_copy,
    "fill": _inherit_copy,
    "fill-rule": _inherit_copy,
    "style": _inherit_copy,
    "transform": _inherit_matrix_multiply,
    "stroke": _inherit_copy,
    "stroke-width": _inherit_copy,
    "stroke-linecap": _inherit_copy,
    "stroke-linejoin": _inherit_copy,
    "stroke-miterlimit": _inherit_copy,
    "stroke-dasharray": _inherit_copy,
    "stroke-dashoffset": _inherit_copy,
    "stroke-opacity": _inherit_copy,
    "fill-opacity": _inherit_copy,
    "opacity": _inherit_multiply,
    "clip-path": _inherit_clip_path,
    "id": _do_not_inherit,
    "data-name": _do_not_inherit,
    "enable-background": _do_not_inherit,
    "overflow": _inherit_nondefault_overflow,
}


_INHERITABLE_ATTRIB = frozenset(
    k for k, v in _INHERIT_ATTRIB_HANDLERS.items() if v is not _do_not_inherit
)

_INHERITABLE_ATTRIB_DEFAULTS = {
    k: (
        ntos(ATTRIB_DEFAULTS[k])
        if isinstance(ATTRIB_DEFAULTS[k], numbers.Number)
        else str(ATTRIB_DEFAULTS[k])
    )
    for k, v in _INHERIT_ATTRIB_HANDLERS.items()
    if k in ATTRIB_DEFAULTS and v is _inherit_copy
}


def _attr_supported(el: etree.Element, attr_name: str) -> bool:
    tag = strip_ns(el.tag)
    field_name = _field_name(attr_name)
    if tag in _VALID_FIELDS:
        return field_name in _VALID_FIELDS[tag]
    return True  # we don't know


def _drop_default_attrib(attrib: MutableMapping[str, Any]):
    for attr_name in sorted(attrib.keys()):
        value = attrib[attr_name]
        default_value = attrib_default(attr_name, default=None)
        if default_value is None:
            continue
        if isinstance(default_value, float):
            value = float(value)
        if default_value == value:
            del attrib[attr_name]


def _inherit_attrib(
    attrib: Mapping[str, Any],
    child: etree.Element,
    skip_unhandled: bool = False,
    skips=frozenset(),
):
    attrib = copy.deepcopy(attrib)
    for attr_name in sorted(attrib.keys()):
        if attr_name in skips or not _attr_supported(child, attr_name):
            del attrib[attr_name]
            continue
        if not attr_name in _INHERIT_ATTRIB_HANDLERS:
            continue
        _INHERIT_ATTRIB_HANDLERS[attr_name](attrib, child, attr_name)
        del attrib[attr_name]

    if len(attrib) and not skip_unhandled:
        raise ValueError(f"Unable to process attrib {attrib}")
